
package de.xoev.xta.test.app.server;

import java.math.BigInteger;
import java.util.Optional;

import org.apache.cxf.annotations.SchemaValidation;
import org.apache.cxf.ws.addressing.AttributedURIType;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import de.xoev.xta.test.app.service.CustomizationService;
import de.xoev.xta.test.app.service.ParameterValidatorService;
import de.xoev.xta.test.app.service.ProtocolService;
import de.xoev.xta.test.app.service.XtaParameter;
import de.xoev.xta.test.app.util.UtilComponent;
import de.xoev.xta.test.app.util.XtaIdGenerator;
import genv3.de.xoev.transport.xta.x211.InvalidMessageIDException;
import genv3.de.xoev.transport.xta.x211.MsgBoxPortType;
import genv3.de.xoev.transport.xta.x211.PermissionDeniedException;
import genv3.de.xoev.transport.xta.x211.XTAWSTechnicalProblemException;
import genv3.eu.osci.ws.x2008.x05.transport.MsgBoxResponseType;
import genv3.eu.osci.ws.x2008.x05.transport.MsgBoxStatusListRequestType;
import genv3.eu.osci.ws.x2008.x05.transport.MsgSelector;
import genv3.eu.osci.ws.x2008.x05.transport.MsgStatusListType;
import genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData;
import jakarta.jws.WebService;
import lombok.extern.log4j.Log4j2;

@WebService(serviceName = "XTAService", portName = "MsgBoxPort", targetNamespace = "http://xoev.de/transport/xta/211", wsdlLocation = "classpath:wsdl/XTA.wsdl", endpointInterface = "genv3.de.xoev.transport.xta.x211.MsgBoxPortType")
@SchemaValidation(type = SchemaValidation.SchemaValidationType.BOTH)
@Component
@Log4j2
public class MsgBoxPortImpl implements MsgBoxPortType {

    protected static final String HEADER_VALUE_NOT_FOUND = "Die erforderliche Header Variable '%s' wurde nicht gefunden";

    @Autowired
    private ParameterValidatorService validator;

    @Autowired
    private CustomizationService customizationService;

    @Autowired
    private ProtocolService protocolService;

    @Autowired
    private UtilComponent utilComponent;

    /*
     * (non-Javadoc)
     * @see genv3.de.xoev.transport.xta.x211.MsgBoxPortType#getMessage(genv3.eu.osci.ws.x2008.x05.transport.
     * MsgBoxFetchRequest fetchRequest, genv3.eu.osci.ws.x2014.x10.transport.PartyType authorIdentifier,
     * genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData messageMetaData,
     * genv3.eu.osci.ws.x2008.x05.transport.MsgBoxResponseType fetchResponseHeader)*
     */
    @Override
    public genv3.de.xoev.transport.xta.x211.GenericContentContainer getMessage(
            final genv3.eu.osci.ws.x2008.x05.transport.MsgBoxFetchRequest fetchRequest,
            final genv3.eu.osci.ws.x2014.x10.transport.PartyType authorIdentifier,
            final jakarta.xml.ws.Holder<genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData> messageMetaData,
            final jakarta.xml.ws.Holder<genv3.eu.osci.ws.x2008.x05.transport.MsgBoxResponseType> fetchResponseHeader)
            throws XTAWSTechnicalProblemException, InvalidMessageIDException, PermissionDeniedException {
        log.info("Executing operation getMessage");
        //
        validator.checkMsgSelector(fetchRequest.getMsgSelector(),
                protocolService.getParameter(XtaParameter.MESSAGE_ID, String.class));
        validator.isNotNull(fetchRequest.getMsgPart(), "MsgPart");
        MessageMetaData messageMetaDataReturn = new MessageMetaData();
        utilComponent.createMessageMetaData(messageMetaDataReturn,
                protocolService.getParameter(XtaParameter.MESSAGE_ID, String.class,
                        () -> XtaIdGenerator.generateMessageId().toString()));

        MsgBoxResponseType msgBoxResponse = new MsgBoxResponseType();
        msgBoxResponse
                .setMsgBoxRequestID(protocolService.getParameter(XtaParameter.MSG_BOX_REQUEST_ID, String.class)
                        .orElseThrow(() -> new XTAWSTechnicalProblemException("MessageID not found")));
        protocolService.addParameter(XtaParameter.MSG_BOX_REQUEST_ID,
                protocolService.getParameter(XtaParameter.MSG_BOX_REQUEST_ID, String.class));

        return customizationService.createGenericContentContainer();
    }

    /*
     * (non-Javadoc)
     * @see genv3.de.xoev.transport.xta.x211.MsgBoxPortType#getStatusList(genv3.eu.osci.ws.x2008.x05.transport.
     * MsgBoxStatusListRequestType fetchRequest, genv3.eu.osci.ws.x2014.x10.transport.PartyType authorIdentifier,
     * genv3.eu.osci.ws.x2008.x05.transport.MsgBoxResponseType fetchResponseHeader)*
     */
    @Override
    public genv3.eu.osci.ws.x2008.x05.transport.MsgStatusListType getStatusList(
            final genv3.eu.osci.ws.x2008.x05.transport.MsgBoxStatusListRequestType fetchRequest,
            final genv3.eu.osci.ws.x2014.x10.transport.PartyType authorIdentifier,
            final jakarta.xml.ws.Holder<genv3.eu.osci.ws.x2008.x05.transport.MsgBoxResponseType> fetchResponseHeader)
            throws XTAWSTechnicalProblemException, PermissionDeniedException {
        log.info("Executing operation getStatusList");
        validator.checkMsgSelector(fetchRequest.getMsgSelector(), Optional.empty());
        validator.isNotNull(fetchRequest.getListForm(), "listForm");
        validator.isNotNull(fetchRequest.getMaxListItems(), "maxListItems");

        // use msg-id from filter request or gen new if there was none
        String messageId = Optional.ofNullable(fetchRequest)
                .map(MsgBoxStatusListRequestType::getMsgSelector)
                .map(MsgSelector::getMessageID)
                .map(list -> {
                    return list.isEmpty() ? null : list.get(0);
                })
                .map(AttributedURIType::getValue)
                .orElseGet(() -> XtaIdGenerator.generateMessageId().toString());

        // save messageID
        protocolService.addParameter(XtaParameter.MESSAGE_ID, messageId);

        MsgStatusListType statusListResponse = new MsgStatusListType();
        MessageMetaData messageMetaData = new MessageMetaData();
        statusListResponse.getMessageMetaData().add(messageMetaData);
        utilComponent.createMessageMetaData(messageMetaData, messageId);

        MsgBoxResponseType msgBoxResponse = new MsgBoxResponseType();
        String msgBoxRequestId = XtaIdGenerator.generateRequestId().toString();
        msgBoxResponse.setMsgBoxRequestID(msgBoxRequestId);
        msgBoxResponse.setItemsPending(BigInteger.ONE);
        msgBoxResponse.setNoMessageAvailable(null);

        fetchResponseHeader.value = msgBoxResponse;
        // save msgBoxRequestID
        protocolService.addParameter(XtaParameter.MSG_BOX_REQUEST_ID, msgBoxRequestId);
        return statusListResponse;
    }

    /*
     * (non-Javadoc)
     * @see genv3.de.xoev.transport.xta.x211.MsgBoxPortType#getNextMessage(genv3.eu.osci.ws.x2008.x05.transport.
     * MsgBoxGetNextRequestType fetchRequest, genv3.eu.osci.ws.x2014.x10.transport.PartyType authorIdentifier,
     * genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData messageMetaData,
     * genv3.eu.osci.ws.x2008.x05.transport.MsgBoxResponseType fetchResponseHeader)*
     */
    @Override
    public genv3.de.xoev.transport.xta.x211.GenericContentContainer getNextMessage(
            final genv3.eu.osci.ws.x2008.x05.transport.MsgBoxGetNextRequestType fetchRequest,
            final genv3.eu.osci.ws.x2014.x10.transport.PartyType authorIdentifier,
            final jakarta.xml.ws.Holder<genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData> messageMetaData,
            final jakarta.xml.ws.Holder<genv3.eu.osci.ws.x2008.x05.transport.MsgBoxResponseType> fetchResponseHeader)
            throws XTAWSTechnicalProblemException, InvalidMessageIDException, PermissionDeniedException {
        log.info("Executing operation getNextMessage");
        throw new XTAWSTechnicalProblemException("Method not supported");
    }

    /*
     * (non-Javadoc)
     * @see genv3.de.xoev.transport.xta.x211.MsgBoxPortType#getNextStatusList(genv3.eu.osci.ws.x2008.x05.transport.
     * MsgBoxGetNextRequestType fetchRequest, genv3.eu.osci.ws.x2014.x10.transport.PartyType authorIdentifier,
     * genv3.eu.osci.ws.x2008.x05.transport.MsgBoxResponseType fetchResponseHeader)*
     */
    @Override
    public genv3.eu.osci.ws.x2008.x05.transport.MsgStatusListType getNextStatusList(
            final genv3.eu.osci.ws.x2008.x05.transport.MsgBoxGetNextRequestType fetchRequest,
            final genv3.eu.osci.ws.x2014.x10.transport.PartyType authorIdentifier,
            final jakarta.xml.ws.Holder<genv3.eu.osci.ws.x2008.x05.transport.MsgBoxResponseType> fetchResponseHeader)
            throws XTAWSTechnicalProblemException, PermissionDeniedException {
        log.info("Executing operation getNextStatusList");
        throw new XTAWSTechnicalProblemException("Method not supported");
    }

    /*
     * (non-Javadoc)
     * @see
     * genv3.de.xoev.transport.xta.x211.MsgBoxPortType#close(genv3.eu.osci.ws.x2008.x05.transport.MsgBoxCloseRequestType
     * fetchRequest, genv3.eu.osci.ws.x2014.x10.transport.PartyType authorIdentifier)*
     */
    @Override
    public void close(final genv3.eu.osci.ws.x2008.x05.transport.MsgBoxCloseRequestType fetchRequest,
            final genv3.eu.osci.ws.x2014.x10.transport.PartyType authorIdentifier)
            throws XTAWSTechnicalProblemException, InvalidMessageIDException, PermissionDeniedException {
        log.info("Executing operation close");
        if (validator.isNotBlank(fetchRequest.getMsgBoxRequestID(), XtaParameter.MSG_BOX_REQUEST_ID)) {
            Optional<String> requestId = protocolService.getParameter(XtaParameter.MSG_BOX_REQUEST_ID, String.class);
            if (requestId.isPresent()) {
                validator.isEquals(requestId.get(), fetchRequest.getMsgBoxRequestID(), XtaParameter.MSG_BOX_REQUEST_ID);
            }
        }
        if (validator.isListSizeEquals(fetchRequest.getLastMsgReceived(), 1, "LastMsgReceived")) {
            validator.isNotBlank(fetchRequest.getLastMsgReceived().get(0).getValue(), XtaParameter.MESSAGE_ID);
            Optional<String> messageId = protocolService.getParameter(XtaParameter.MESSAGE_ID, String.class);
            if (messageId.isPresent()) {
                validator.isEquals(messageId.get(), fetchRequest.getLastMsgReceived().get(0).getValue(),
                        XtaParameter.MESSAGE_ID);
            }
        }
    }

}
