
package de.xoev.xta.test.app.server;

import java.util.Optional;

import org.apache.cxf.annotations.SchemaValidation;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import de.xoev.xta.test.app.service.CustomizationService;
import de.xoev.xta.test.app.service.ParameterValidatorService;
import de.xoev.xta.test.app.service.ProtocolService;
import de.xoev.xta.test.app.service.XtaParameter;
import de.xoev.xta.test.app.util.UtilComponent;
import de.xoev.xta.test.app.util.XtaIdGenerator;
import genv3.de.xoev.transport.xta.x211.GenericContentContainer;
import genv3.de.xoev.transport.xta.x211.MessageSchemaViolationException;
import genv3.de.xoev.transport.xta.x211.MessageVirusDetectionException;
import genv3.de.xoev.transport.xta.x211.ParameterIsNotValidException;
import genv3.de.xoev.transport.xta.x211.PermissionDeniedException;
import genv3.de.xoev.transport.xta.x211.SendPortType;
import genv3.de.xoev.transport.xta.x211.SyncAsyncException;
import genv3.de.xoev.transport.xta.x211.XTAWSTechnicalProblemException;
import genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData;
import jakarta.jws.WebService;
import lombok.extern.log4j.Log4j2;

@WebService(serviceName = "XTAService", portName = "SendXtaPort", targetNamespace = "http://xoev.de/transport/xta/211", wsdlLocation = "classpath:wsdl/XTA.wsdl", endpointInterface = "genv3.de.xoev.transport.xta.x211.SendPortType")
@SchemaValidation(type = SchemaValidation.SchemaValidationType.BOTH)
@Component
@Log4j2
public class SendXtaPortImpl implements SendPortType {

    protected static final String HEADER_VALUE_NOT_FOUND = "Die erforderliche Header Variable '%s' wurde nicht gefunden";

    @Autowired
    private ParameterValidatorService validator;

    @Autowired
    private CustomizationService customizationService;

    @Autowired
    private ProtocolService protocolService;

    @Autowired
    private UtilComponent utilComponent;

    /*
     * (non-Javadoc)
     * @see genv3.de.xoev.transport.xta.x211.SendPortType#sendMessage(genv3.de.xoev.transport.xta.x211.
     * GenericContentContainer genericContainer, genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData messageMetaData,
     * genv3.eu.osci.ws.x2008.x05.transport.X509TokenContainerType x509TokenContainer)*
     */
    @Override
    public void sendMessage(final GenericContentContainer genericContainer,
            final genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData messageMetaData,
            final genv3.eu.osci.ws.x2008.x05.transport.X509TokenContainerType x509TokenContainer)
            throws SyncAsyncException, XTAWSTechnicalProblemException, MessageVirusDetectionException,
            MessageSchemaViolationException, PermissionDeniedException, ParameterIsNotValidException {
        log.info("Executing operation sendMessage");
        if (messageMetaData != null) {
            Optional<String> messageId = protocolService.getParameter(XtaParameter.MESSAGE_ID, String.class);
            validator.checkMessageMetaData(messageMetaData, messageId);
        } else {
            protocolService
                    .addErrorResultMessage(String.format(HEADER_VALUE_NOT_FOUND, XtaParameter.MESSAGE_META_DATA));
        }
        validator.checkMessage(genericContainer);

    }

    /*
     * (non-Javadoc)
     * @see genv3.de.xoev.transport.xta.x211.SendPortType#sendMessageSync(genv3.de.xoev.transport.xta.x211.
     * GenericContentContainer genericContainer, genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData messageMetaData,
     * genv3.eu.osci.ws.x2008.x05.transport.X509TokenContainerType x509TokenContainer)*
     */
    @Override
    public void sendMessageSync(final jakarta.xml.ws.Holder<GenericContentContainer> genericContainer,
            final jakarta.xml.ws.Holder<genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData> messageMetaData,
            final jakarta.xml.ws.Holder<genv3.eu.osci.ws.x2008.x05.transport.X509TokenContainerType> x509TokenContainer)
            throws SyncAsyncException, XTAWSTechnicalProblemException, MessageVirusDetectionException,
            MessageSchemaViolationException, PermissionDeniedException, ParameterIsNotValidException {
        log.info("Executing operation sendMessageSync");
        // messageMetaData is in/out, so we need to use the data first for input and create a new object for output
        Optional<String> messageId = protocolService.getParameter(XtaParameter.MESSAGE_ID, String.class);
        if (messageMetaData != null && messageMetaData.value != null) {
            validator.checkMessageMetaData(messageMetaData.value, messageId);
            messageId = UtilComponent.extractAndSaveMessageID(messageMetaData.value, messageId);
        } else {
            protocolService
                    .addErrorResultMessage(String.format(HEADER_VALUE_NOT_FOUND, XtaParameter.MESSAGE_META_DATA));
        }
        validator.checkMessage(genericContainer.value);

        GenericContentContainer genericContentContainerReturn = customizationService
                .createGenericContentContainer();
        MessageMetaData messageMetadataReturn = new MessageMetaData();

        String messageIdReturn = null;
        if (messageId.isEmpty()) {
            // nothing set in protocol
            messageIdReturn = XtaIdGenerator.generateMessageId().toString();
            protocolService.addParameter(XtaParameter.MESSAGE_ID, messageIdReturn);
        } else {
            messageIdReturn = messageId.get();
        }
        utilComponent.createMessageMetaData(messageMetadataReturn, messageIdReturn);

        messageMetaData.value = messageMetadataReturn;
        genericContainer.value = genericContentContainerReturn;
    }

}
