/*
 * @formatter:off
*
* Copyright 2021-2022  Koordinierungsstelle für IT-Standards (KoSIT)
*
* Licensed under the European Public License, Version 1.2 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     https://opensource.org/licenses/EUPL-1.2
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
* @formatter:on
 */
package de.xoev.xta.test.app.service;

import java.io.IOException;
import java.math.BigInteger;
import java.security.KeyManagementException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.util.HashSet;
import java.util.Optional;

import javax.net.ssl.HostnameVerifier;
import javax.xml.namespace.QName;

import org.apache.commons.lang3.StringUtils;
import org.apache.cxf.bus.spring.SpringBus;
import org.apache.cxf.configuration.jsse.TLSClientParameters;
import org.apache.cxf.transport.http.HTTPConduitConfigurer;
import org.apache.cxf.transports.http.configuration.HTTPClientPolicy;
import org.apache.cxf.ws.addressing.AttributedURIType;
import org.apache.cxf.ws.policy.IgnorablePolicyInterceptorProvider;
import org.apache.cxf.ws.policy.PolicyInterceptorProviderRegistry;
import org.apache.hc.client5.http.ssl.DefaultHostnameVerifier;
import org.apache.hc.client5.http.ssl.NoopHostnameVerifier;
import org.apache.hc.core5.ssl.SSLContextBuilder;
import org.apache.hc.core5.ssl.SSLContexts;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import de.xoev.xta.test.app.exception.XtaTesterException;
import de.xoev.xta.test.app.model.XtaConstants;
import de.xoev.xta.test.app.util.UtilComponent;
import de.xoev.xta.test.app.util.XtaIdGenerator;
import genv3.de.xoev.transport.xta.x211.ContentType;
import genv3.de.xoev.transport.xta.x211.GenericContentContainer;
import genv3.de.xoev.transport.xta.x211.GenericContentContainer.ContentContainer;
import genv3.de.xoev.transport.xta.x211.LookupServiceRequest;
import genv3.de.xoev.transport.xta.x211.LookupServiceRequest.LookupServiceRequestList;
import genv3.de.xoev.transport.xta.x211.LookupServiceResponse;
import genv3.de.xoev.transport.xta.x211.TransportReport;
import genv3.eu.osci.ws.x2008.x05.transport.MsgBoxCloseRequestType;
import genv3.eu.osci.ws.x2008.x05.transport.MsgBoxFetchRequest;
import genv3.eu.osci.ws.x2008.x05.transport.MsgBoxResponseType;
import genv3.eu.osci.ws.x2008.x05.transport.MsgBoxStatusListRequestType;
import genv3.eu.osci.ws.x2008.x05.transport.MsgStatusListType;
import genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData;
import jakarta.annotation.PostConstruct;
import jakarta.xml.ws.Holder;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Service
public class XtaClient {

    @Autowired
    private ConfigurationService configurationService;

    @Autowired
    private ProtocolService protocolService;

    @Autowired
    private ParameterValidatorService validator;

    @Autowired
    private CustomizationService customizationService;

    @Autowired
    private ClientConfigurationContainer clientConfigurationContainer;

    @Autowired
    private UtilComponent utilComponent;

    @Autowired
    private SpringBus bus;

    @PostConstruct
    void setUp() {
        bus.setExtension((name, address, httpConduit) -> {
            try {
                httpConduit.setTlsClientParameters(createTLSClientParameters());
                httpConduit.setClient(getClientPolicy());
            } catch (XtaTesterException e) {
                log.error("exception during setup of spring bus", e);
            }
        }, HTTPConduitConfigurer.class);
        // TODO besser wäre es, diese security policy korrekt zu implementieren statt zu ignorieren
        final var reg = bus.getExtension(PolicyInterceptorProviderRegistry.class);
        final var set = new HashSet<QName>();
        set.add(new QName("http://docs.oasis-open.org/ws-sx/ws-securitypolicy/200702", "Wss11"));
        reg.register(new IgnorablePolicyInterceptorProvider(set));
    }

    /*
     * methods to call the WebServices
     */
    public void checkAccountActive() {
        log.info("Requesting checkAccountActive");
        try {
            clientConfigurationContainer.getManagementPortClient().checkAccountActive(
                    utilComponent.createPartyTypeForAuthor());
            protocolService.addMethodCallSendMessage();
        } catch (Exception e) {
            logProtocolErrorWithoutCrashing(e);
        }
    }

    public void lookupService() {
        log.info("Requesting lookupService");
        try {
            LookupServiceRequest lookupServiceRequest = new LookupServiceRequest();
            LookupServiceRequestList lookupServiceRequestList = new LookupServiceRequestList();
            lookupServiceRequest.getLookupServiceRequestList().add(lookupServiceRequestList);

            lookupServiceRequestList.getLookupService().setReader(utilComponent.createPartyTypeForReader());
            lookupServiceRequestList.getLookupService().setServiceType(
                    StringUtils.trimToNull(configurationService.getXtaConfig().getLookupServiceUri()));

            LookupServiceResponse sendResponse = clientConfigurationContainer.getManagementPortClient()
                    .lookupService(lookupServiceRequest, utilComponent.createPartyTypeForAuthor());

            sendResponse.getLookupServiceResultList().getLookupServiceResult().forEach(
                    e -> log.info("Service {} is available {}", e.getServiceType(), e.getIsServiceAvailableValue()));

        } catch (Exception e) {
            logProtocolErrorWithoutCrashing(e);
        }

    }

    public void createMessageId() {
        log.info("Requesting createMessageId");
        try {
            AttributedURIType response = clientConfigurationContainer.getManagementPortClient()
                    .createMessageId(utilComponent.createPartyTypeForAuthor());
            if (validator.isNotBlank(response.getValue(), "messageId")) {
                protocolService.addParameter(XtaParameter.MESSAGE_ID, response.getValue());
            }
        } catch (Exception e) {
            logProtocolErrorWithoutCrashing(e);
        }
    }

    public void sendMessage() {
        log.info("Requesting sendMessage");

        try {
            GenericContentContainer genericContentContainer = customizationService
                    .createGenericContentContainer();
            String messageId = protocolService.getParameter(XtaParameter.MESSAGE_ID, String.class,
                    () -> XtaIdGenerator.generateMessageId().toString());

            MessageMetaData messageMetaData = new MessageMetaData();
            utilComponent.createMessageMetaData(messageMetaData, messageId);
            clientConfigurationContainer.getSendPortClient().sendMessage(genericContentContainer, messageMetaData,
                    null);

        } catch (Exception e) {
            logProtocolErrorWithoutCrashing(e);
        }
    }

    public void sendMessageSync() {
        log.info("Requesting sendMessageSync");
        try {
            GenericContentContainer genericContentContainer = customizationService
                    .createGenericContentContainer();
            Holder<GenericContentContainer> genericContentContainerHolder = new Holder<>(
                    genericContentContainer);

            String messageId = protocolService.getParameter(XtaParameter.MESSAGE_ID, String.class,
                    () -> XtaIdGenerator.generateMessageId().toString());

            MessageMetaData messageMetaData = new MessageMetaData();
            utilComponent.createMessageMetaData(messageMetaData, messageId);
            Holder<MessageMetaData> messageMetaDataHolder = new Holder<>(messageMetaData);

            clientConfigurationContainer.getSendPortClient().sendMessageSync(genericContentContainerHolder,
                    messageMetaDataHolder, null);

            protocolService.addMethodCallSendMessage();

            // if all was ok, now the holders contain the returned data, if not, we got an exception
            validator.checkMessage(genericContentContainerHolder.value);

            if (messageMetaDataHolder.value != null) {
                validator.checkMessageMetaData(messageMetaDataHolder.value,
                        Optional.of(messageId));
            } else {
                protocolService.addErrorResultMessage(
                        XtaConstants.MESSAGE_META_DATA.getLocalPart() + " nicht im Soap Header angegeben");
            }
        } catch (Exception e) {
            logProtocolErrorWithoutCrashing(e);
        }
    }

    public void getTransportReport() {
        log.info("Requesting getTransportReport");
        try {
            String messageId = protocolService.getParameter(XtaParameter.MESSAGE_ID, String.class,
                    () -> XtaIdGenerator.generateMessageId().toString());
            AttributedURIType attUriType = new AttributedURIType();
            attUriType.setValue(messageId);

            TransportReport transportReport = clientConfigurationContainer.getManagementPortClient().getTransportReport(
                    attUriType, utilComponent.createPartyTypeForAuthor());

            validator.isNotNull(transportReport.getReportTime(), "ReportTime");
            validator.isNotBlank(transportReport.getXTAServerIdentity(), "XTAServerIdentity");
            validator.checkMessageMetaData(transportReport.getMessageMetaData(), Optional.of(messageId));
            if (validator.isNotNull(transportReport.getMessageStatus(), "MessageStatus")) {
                validator.isNotNull(transportReport.getMessageStatus().getStatus(), "MessageStatus.status");
            }
            validator.isNotNull(transportReport.getAdditionalReports(), "AdditionalReports");

        } catch (Exception e) {
            logProtocolErrorWithoutCrashing(e);
        }
    }

    public void getStatusList() {
        log.info("Requesting getStatusList");
        try {
            MsgBoxStatusListRequestType request = new MsgBoxStatusListRequestType();
            Holder<MsgBoxResponseType> responseHolder = new Holder<>();

            request.setMsgSelector(UtilComponent.createMsgSelector(Optional.empty()));
            request.setListForm("MessageMetaData");
            request.setMaxListItems(BigInteger.ONE);
            MsgStatusListType statusListResponse = clientConfigurationContainer.getMsgBoxPortClient().getStatusList(
                    request, utilComponent.createPartyTypeForAuthor(), responseHolder);

            handleGetStatusListResponse(statusListResponse, responseHolder);
        } catch (Exception e) {
            logProtocolErrorWithoutCrashing(e);
        }
    }

    private void handleGetStatusListResponse(final MsgStatusListType statusListResponse,
            final Holder<MsgBoxResponseType> responseHolder) {
        if (validator.isListNotEmpty(statusListResponse.getMessageMetaData(), "MessageMetaData")) {
            MessageMetaData messageMetaData = statusListResponse.getMessageMetaData().get(0);

            validator.checkMessageMetaData(messageMetaData, Optional.empty());
            if (messageMetaData.getMsgIdentification() != null
                    && messageMetaData.getMsgIdentification().getMessageID() != null) {
                protocolService.addParameter(XtaParameter.MESSAGE_ID,
                        messageMetaData.getMsgIdentification().getMessageID().getValue());
            }
        }

        MsgBoxResponseType msgBoxResponseType = responseHolder.value;
        if (msgBoxResponseType != null) {
            if (validator.isNotBlank(msgBoxResponseType.getMsgBoxRequestID(),
                    XtaParameter.MSG_BOX_RESPONSE.join("MsgBoxRequestID"))) {
                protocolService.addParameter(XtaParameter.MSG_BOX_REQUEST_ID, msgBoxResponseType.getMsgBoxRequestID());
            }

            if (msgBoxResponseType.getNoMessageAvailable().getReason() != null) {
                protocolService.addErrorResultMessage("No message available");
            }
            validator.isNotNull(msgBoxResponseType.getItemsPending(),
                    XtaParameter.MSG_BOX_RESPONSE.join("ItemsPending"));
        } else {
            protocolService.addErrorResultMessage(
                    XtaConstants.MSG_BOX_RESPONSE.getLocalPart() + " nicht im Soap Header angegeben");
        }
    }

    public void getMessage() {
        log.info("Requesting getMessage");

        MsgBoxFetchRequest request = new MsgBoxFetchRequest();

        request.setMsgSelector(
                UtilComponent.createMsgSelector(protocolService.getParameter(XtaParameter.MESSAGE_ID, String.class)));
        request.setMsgPart("Body");

        Holder<MessageMetaData> messageMetaDataHolder = new Holder<>();
        Holder<MsgBoxResponseType> msgBoxResponseHolder = new Holder<>();
        try {
            GenericContentContainer containerReturn = clientConfigurationContainer.getMsgBoxPortClient().getMessage(
                    request, utilComponent.createPartyTypeForAuthor(), messageMetaDataHolder, msgBoxResponseHolder);
            checkMessage(containerReturn);
        } catch (Exception e) {
            logProtocolErrorWithoutCrashing(e);
        }
    }

    public void close() {
        log.info("Requesting close");

        MsgBoxCloseRequestType request = new MsgBoxCloseRequestType();
        String messageId = protocolService.getParameter(XtaParameter.MESSAGE_ID, String.class,
                () -> XtaIdGenerator.generateMessageId().toString());

        String msgBoxRequestId = protocolService.getParameter(XtaParameter.MSG_BOX_REQUEST_ID, String.class,
                () -> XtaIdGenerator.generateRequestId().toString());

        request.setMsgBoxRequestID(msgBoxRequestId);
        AttributedURIType attUriType = new AttributedURIType();
        attUriType.setValue(messageId);
        request.getLastMsgReceived().add(attUriType);

        try {

            clientConfigurationContainer.getMsgBoxPortClient().close(request, utilComponent.createPartyTypeForAuthor());
        } catch (Exception e) {
            logProtocolErrorWithoutCrashing(e);
        }
    }

    private void logProtocolErrorWithoutCrashing(Exception e) {
        log.error("suppressed exception, adding to protocol", e);
        protocolService.addErrorResultMessage(e.getMessage());
    }

    private void checkMessage(final GenericContentContainer request) {
        if (request.getContentContainer() == null && request.getEncryptedData() == null) {
            log.error("No ContentContainer or EncryptedData element found");
            protocolService.addErrorResultMessage("Keine Nachricht enthalten");
        } else if (request.getContentContainer() != null) {
            ContentContainer contentContainer = request.getContentContainer();
            if (validator.isNotNull(contentContainer.getMessage(), "message")) {
                ContentType message = contentContainer.getMessage();
                validator.isNotNull(message.getValue(), "message.value");
                validator.isNotBlank(message.getContentType(), "contentType");
                validator.isNotBlank(message.getEncoding(), "encoding");
                validator.isNotBlank(message.getFilename(), "filename");
                validator.isNotBlank(message.getId(), "id");
                validator.isNotNull(message.getSize(), "size");
            }
        } else {
            protocolService.addSuccessResultMessage("Encrypted Message received");
        }
    }

    private HTTPClientPolicy getClientPolicy() {
        // set timeouts
        HTTPClientPolicy clientPolicy = new HTTPClientPolicy();
        clientPolicy.setConnectionTimeout(30000);
        clientPolicy.setReceiveTimeout(30000);
        clientPolicy.setAutoRedirect(true);
        clientPolicy.setConnectionRequestTimeout(30000);
        clientPolicy.setAllowChunking(true);
        return clientPolicy;
    }

    private TLSClientParameters createTLSClientParameters() throws XtaTesterException {
        TLSClientParameters tlsClientParameters = new TLSClientParameters();
        try {
            SSLContextBuilder sslBuilder = SSLContexts.custom();
            sslBuilder.loadTrustMaterial(configurationService.getClientProperties().getTrustStoreAsResource().getURL(),
                    configurationService.getClientProperties().getTrustStorePassword().toCharArray());
            sslBuilder.loadKeyMaterial(configurationService.getClientProperties().getKeyStoreAsResource().getURL(),
                    configurationService.getClientProperties().getKeyStorePassword().toCharArray(),
                    configurationService.getClientProperties().getKeyPassword().toCharArray(),
                    (aliases, sslParameters) -> {
                        String alias = configurationService.getClientProperties().getKeyAlias();
                        if ((alias == null || alias.isBlank()) && aliases.size() == 1) {
                            // use the first alias from the map
                            alias = aliases.keySet().iterator().next();
                        }
                        return alias;
                    });
            HostnameVerifier hostnameverifier = configurationService.getClientProperties()
                    .isCheckHostnameInCertificate()
                            ? new DefaultHostnameVerifier()
                            : new NoopHostnameVerifier();

            tlsClientParameters.setHostnameVerifier(hostnameverifier);

            // only by setting socket Factory, this ssl configuration works with cxf 4
            tlsClientParameters.setSSLSocketFactory(sslBuilder.build().getSocketFactory());
        } catch (NoSuchAlgorithmException | KeyStoreException | CertificateException | IOException
                | KeyManagementException | UnrecoverableKeyException e) {
            throw new XtaTesterException("Unable to create Soap Client", e);
        }
        return tlsClientParameters;
    }
}