package de.xoev.xta.test.app.util;

import java.math.BigInteger;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Optional;

import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;

import org.apache.commons.lang3.StringUtils;
import org.apache.cxf.ws.addressing.AttributedURIType;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import de.xoev.xta.test.app.model.Role;
import de.xoev.xta.test.app.service.ConfigurationService;
import de.xoev.xta.test.app.service.ProtocolService;
import de.xoev.xta.test.app.service.XtaConfigDtoConverter;
import genv3.eu.osci.ws.x2008.x05.transport.MsgSelector;
import genv3.eu.osci.ws.x2014.x10.transport.DeliveryAttributesType;
import genv3.eu.osci.ws.x2014.x10.transport.DestinationsType;
import genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData;
import genv3.eu.osci.ws.x2014.x10.transport.MsgIdentificationType;
import genv3.eu.osci.ws.x2014.x10.transport.OriginatorsType;
import genv3.eu.osci.ws.x2014.x10.transport.PartyType;
import genv3.eu.osci.ws.x2014.x10.transport.QualifierType;

@Component
public class UtilComponent {

    @Autowired
    private ConfigurationService configurationService;

    @Autowired
    private XtaConfigDtoConverter xtaConfigDtoConverter;

    @Autowired
    private ProtocolService protocolService;

    public static Optional<String> extractAndSaveMessageID(final MessageMetaData messageMetaData,
            final Optional<String> savedMessageId) {
        if (savedMessageId.isPresent()) {
            return savedMessageId;
        }
        if (messageMetaData.getMsgIdentification() != null) {
            return Optional.ofNullable(messageMetaData.getMsgIdentification().getMessageID().getValue());
        }
        return Optional.empty();
    }

    public static XMLGregorianCalendar createGC(final int offsetMonth) {

        GregorianCalendar cal = new GregorianCalendar();
        if (offsetMonth != 0) {
            cal.add(Calendar.MONTH, offsetMonth);
        }
        return DatatypeFactory.newDefaultInstance().newXMLGregorianCalendar(cal);
    }

    public static MsgSelector createMsgSelector(final Optional<String> messageId) {
        MsgSelector msgSelector = new MsgSelector();
        msgSelector.setNewEntry(Boolean.TRUE);
        if (messageId.isPresent()) {
            AttributedURIType attUriType = new AttributedURIType();
            attUriType.setValue(messageId.get());
            msgSelector.getMessageID().add(attUriType);
        }
        msgSelector.setMsgBoxEntryTimeFrom(createGC(-1));
        msgSelector.setMsgBoxEntryTimeTo(createGC(0));
        return msgSelector;
    }

    public void createMessageMetaData(final MessageMetaData messageMetaData, final String messageId) {
        XMLGregorianCalendar calendarNow = UtilComponent.createGC(0);

        DeliveryAttributesType deliveryAttributes = new DeliveryAttributesType();
        messageMetaData.setDeliveryAttributes(deliveryAttributes);
        final var role = protocolService.getCurrentScenarioDefinition()
                .map(it -> it.getScenarioRole().getRolePerformedByTestumgebung())
                .orElse(Role.LESER);
        switch (role) {
            case LESER:
                // fallthrough
                deliveryAttributes.setDelivery(calendarNow);
            case EMPFAENGER:
                // fallthrough
                deliveryAttributes.setInitialFetch(calendarNow);
            case SENDER:
                // fallthrough
                deliveryAttributes.setInitialSend(calendarNow);
            case AUTOR:
                deliveryAttributes.setOrigin(calendarNow);
                break;
            default:
                throw new RuntimeException("role not supported");
        }

        deliveryAttributes.setServiceQuality(StringUtils.trimToNull(
                configurationService.getXtaConfig().getDeliveryAttributesServiceQuality()));

        OriginatorsType originators = new OriginatorsType();
        messageMetaData.setOriginators(originators);

        originators.setAuthor(createPartyTypeForAuthor());
        if (role != Role.AUTOR) {
            originators.setSender(createPartyTypeForSender());
        }

        DestinationsType destination = new DestinationsType();
        messageMetaData.setDestinations(destination);
        destination.setReader(createPartyTypeForReader());

        MsgIdentificationType msgIdentification = new MsgIdentificationType();
        messageMetaData.setMsgIdentification(msgIdentification);
        AttributedURIType attUriType = new AttributedURIType();
        attUriType.setValue(messageId);
        msgIdentification.setMessageID(attUriType);

        messageMetaData.setQualifier(createQualifierType());
        messageMetaData.setMsgSize(BigInteger.valueOf(10L));
    }

    public PartyType createPartyTypeForAuthor() {
        return xtaConfigDtoConverter.fromModel(configurationService.getXtaConfig().getAuthorIdentifierConfig());
    }

    public PartyType createPartyTypeForSender() {
        return xtaConfigDtoConverter.fromModel(configurationService.getXtaConfig().getSenderIdentifierConfig());
    }

    public PartyType createPartyTypeForReader() {
        return xtaConfigDtoConverter.fromModel(configurationService.getXtaConfig().getReaderIdentifierConfig());
    }

    public QualifierType createQualifierType() {
        return xtaConfigDtoConverter.qualifierFromConfig(configurationService.getXtaConfig().getQualifierConfig());
    }

}
